from django.test import TestCase
from django.contrib.auth import get_user_model, authenticate

ModernUser = get_user_model()


class ModernUserManagerTestCase(TestCase):
    def test_create_user_no_email(self):
        with self.assertRaises(ValueError) as context:
            ModernUser.objects._create_user(None, "password")
        self.assertTrue("Users must have an email address." in str(context.exception))

    def test_create_user_with_email(self):
        user = ModernUser.objects.create_user(
            email="john.doe@example.com", password="testpassword123"
        )
        self.assertEqual(user.email, "john.doe@example.com")
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

    def test_create_superuser_is_staff_false(self):
        with self.assertRaises(ValueError) as context:
            ModernUser.objects.create_superuser(
                email="john.doe@example.com", password="testpassword123", is_staff=False
            )
        self.assertTrue("Superuser must have is_staff=True." in str(context.exception))

    def test_create_superuser_is_superuser_false(self):
        with self.assertRaises(ValueError) as context:
            ModernUser.objects.create_superuser(
                email="john.doe@example.com",
                password="testpassword123",
                is_superuser=False,
            )
        self.assertTrue(
            "Superuser must have is_superuser=True." in str(context.exception)
        )

    def test_create_superuser(self):
        user = ModernUser.objects.create_superuser(
            email="john.doe@example.com", password="testpassword123"
        )
        self.assertEqual(user.email, "john.doe@example.com")
        self.assertTrue(user.is_staff)
        self.assertTrue(user.is_superuser)


class ModernUserTests(TestCase):
    def test_absence_of_default_fields(self):
        user = ModernUser.objects.create_user(
            email="test@example.com", password="testpassword123"
        )
        self.assertIsNone(user.username)
        self.assertIsNone(user.first_name)
        self.assertIsNone(user.last_name)

    def test_lower_case_email(self):
        email = "Test@ExAMPle.com"
        user = ModernUser.objects.create_user(email=email, password="testpassword123")
        self.assertEqual(user.email, email.lower())

    def test_create_regular_user(self):
        user = ModernUser.objects.create_user(
            email="user@example.com", password="userpassword123"
        )
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)

    def test_create_superuser(self):
        user = ModernUser.objects.create_superuser(
            email="admin@example.com", password="adminpassword123"
        )
        self.assertTrue(user.is_staff)
        self.assertTrue(user.is_superuser)

    def test_email_uniqueness(self):
        ModernUser.objects.create_user(
            email="unique@example.com", password="testpassword123"
        )
        with self.assertRaises(Exception):
            ModernUser.objects.create_user(
                email="unique@example.com", password="anotherpassword123"
            )
        # Testing case insensitivity
        with self.assertRaises(Exception):
            ModernUser.objects.create_user(
                email="UNIQUE@example.com", password="anotherpassword123"
            )

    def test_user_authentication(self):
        email = "auth@example.com"
        password = "testpassword123"
        ModernUser.objects.create_user(email=email, password=password)
        self.assertIsNotNone(authenticate(email=email, password=password))

    def test_user_authentication_email_case_insensitive(self):
        email = "auth@example.com"
        password = "testpassword123"
        # Test with a capitalized email
        ModernUser.objects.create_user(email=email, password=password)
        self.assertIsNotNone(authenticate(email=email.upper(), password=password))

    def test_required_fields(self):
        self.assertEqual(ModernUser.REQUIRED_FIELDS, [])

    def test_user_str(self):
        """Test the __str__ method of the ModernUser model."""
        user = ModernUser.objects.create_user(
            email="test@example.com", password="testpassword123"
        )
        self.assertEqual(user.__str__(), "test@example.com")



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from .models import ModernUser


@admin.register(ModernUser)
class ModernUserAdmin(UserAdmin):
    """Define admin model for custom User model with no email field.

    You can find detailed examples here:
    https://docs.djangoproject.com/en/3.1/topics/auth/customizing/#a-full-example
    https://www.fomfus.com/articles/how-to-use-email-as-username-for-django-authentication-removing-the-username#Register%20your%20new%20User%20model%20with%20Django%20admin
    """

    fieldsets = (
        (None, {"fields": ("email", "password")}),
        (
            "Permissions",
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                )
            },
        ),
        ("Important dates", {"fields": ("last_login", "date_joined")}),
    )
    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("email", "password1", "password2"),
            },
        ),
    )
    list_display = ("email", "is_staff")
    search_fields = ("email",)
    ordering = ("email",)



from django.db import models
from django.contrib.auth.models import AbstractUser, BaseUserManager


class ModernUserManager(BaseUserManager):
    """A custum user manager that accounts for the modified fields of the custom User model.

    "Writing a manager for a custom user model" (Django docs):
    https://docs.djangoproject.com/en/3.1/topics/auth/customizing/#writing-a-manager-for-a-custom-user-model
    """

    # Serialize the manager into migrations
    # https://docs.djangoproject.com/en/3.1/topics/migrations/#model-managers
    use_in_migrations = True

    def _create_user(self, email, password, **extra_fields):
        """Perform the steps needed to create any kind of User."""
        if not email:
            raise ValueError("Users must have an email address.")
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save()
        return user

    def create_user(self, email, password=None, **extra_fields):
        """Perform the unique steps needed to create a normal user."""
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        """Perform the unique steps needed to create a superuser."""
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return self._create_user(email, password, **extra_fields)

    def get_by_natural_key(self, username):
        """
        Retrieves a user instance using the contents
        of the field nominated by USERNAME_FIELD.

        In `django-modern-user`, the USERNAME_FIELD is the email field.
        We want to ensure that when a user is looked up by their email,
        the lookup is case-insensitive.
        """
        email = username.lower()
        return super().get_by_natural_key(email)


class ModernUser(AbstractUser):
    """A custom user model for this project.

    Django highly recommends setting up a custom User model, even if the default user model is sufficient.
    https://docs.djangoproject.com/en/3.1/topics/auth/customizing/#using-a-custom-user-model-when-starting-a-project
    """

    # Remove some of the default fields.
    username = None
    first_name = None
    last_name = None

    # The email field should be required and unique (unlike default).
    email = models.EmailField("email address", unique=True)

    USERNAME_FIELD = "email"

    # Email is automatically required, since it is the USERNAME_FIELD.
    REQUIRED_FIELDS = []

    # Set the custom user manager
    objects = ModernUserManager()

    def save(self, *args, **kwargs):
        # Convert the email to lowercase before saving
        self.email = self.email.lower()
        super().save(*args, **kwargs)
